/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "ServiceCommunicator.h"
#include "MixCast.h"
#include "Tracking/TrackingUtil.h"
#include "Transmission/ServiceMutexThread.h"
#include "Utility/ProjectUtils.hpp"
#include "Utility/StringUtils.hpp"

FServiceCommunicator::FServiceCommunicator()
	: ServiceSDKHandler(std::make_shared<FServiceSDKHandler>())
	, ServiceSDKServer("SERVICE_SDK", ServiceSDKHandler)
	, SDKServiceClient("SDK_SERVICE")
	, SharedTextureClient("SHAREDTEXTURECOMMUNICATION")
	, ServiceMutexCheck(nullptr)
{
	ServiceSDKHandler->OnServiceStarted().AddRaw(this, &FServiceCommunicator::NotifySdkStarted);

	ActiveChangedDelegate = ((FMixCast*)&IMixCast::Get())->OnActiveChanged().AddRaw(this, &FServiceCommunicator::HandleActiveChanged);

	NotifySdkStarted();
}

FServiceCommunicator::~FServiceCommunicator()
{
	NotifySdkStopped();

	if(IMixCast::IsAvailable())
		((FMixCast*)&IMixCast::Get())->OnActiveChanged().Remove(ActiveChangedDelegate);

	if (ServiceMutexCheck) {
		delete ServiceMutexCheck;
		ServiceMutexCheck = nullptr;
	}
}


bool FServiceCommunicator::SendExperienceEvent(const FString& EventId) const
{
	try
	{
		SDKServiceClient.Run<void>([&EventId](SDK_ServiceClient& Client)
		{
			const std::string EventIdString(TCHAR_TO_UTF8(*EventId));
			Client.SendExperienceEvent(EventIdString);
		});

		MIXCAST_LOG("Successfully sent experience event: %s", *EventId);
		return true;
	}
	catch (...)
	{
		MIXCAST_LOG("Failed to send experience event: %s", *EventId);
		return false;
	}
}

bool FServiceCommunicator::SetAutoSnapshot(const bool bEnable, const FString& CameraId) const
{
	try
	{
		SDKServiceClient.Run<void>([bEnable, &CameraId](SDK_ServiceClient& Client)
		{
			const std::string CameraIdString(TCHAR_TO_UTF8(*CameraId));
			Client.RequestSetAutoSnapshot(CameraIdString, bEnable);
		});

		MIXCAST_LOG("Successfully set auto snapshot for camera: %s", *CameraId);
		return true;
	}
	catch (...)
	{
		MIXCAST_LOG("Failed to set auto snapshot for camera: %s", *CameraId);
		return false;
	}
}

bool FServiceCommunicator::SetVideoRecording(const bool bEnable, const FString& CameraId) const
{
	try
	{
		SDKServiceClient.Run<void>([bEnable, &CameraId](SDK_ServiceClient& Client)
		{
			const std::string CameraIdString(TCHAR_TO_UTF8(*CameraId));
			Client.RequestSetVideoRecording(CameraIdString, bEnable);
		});

		MIXCAST_LOG("Set video recording for camera: %s", *CameraId);
		return true;
	}
	catch (...)
	{
		MIXCAST_LOG("Failed to set video recording for camera: %s", *CameraId);
		return false;
	}
}

bool FServiceCommunicator::SetVideoStreaming(const bool bEnable, const FString& CameraId) const
{
	try
	{
		SDKServiceClient.Run<void>([bEnable, &CameraId](SDK_ServiceClient& Client)
		{
			const std::string CameraIdString(TCHAR_TO_UTF8(*CameraId));
			Client.RequestSetVideoStreaming(CameraIdString, bEnable);
		});

		MIXCAST_LOG("Successfully set video streaming for camera: %s", *CameraId);
		return true;
	}
	catch (...)
	{
		MIXCAST_LOG("Failed to set video streaming for camera: %s", *CameraId);
		return false;
	}
}

bool FServiceCommunicator::TakeSnapshot(const FString& CameraId) const
{
	try
	{
		SDKServiceClient.Run<void>([&CameraId](SDK_ServiceClient& Client)
		{
			const std::string CameraIdString(TCHAR_TO_UTF8(*CameraId));
			Client.RequestTakeSnapshot(CameraIdString);
		});

		MIXCAST_LOG("Successfully took snapshot for camera: %s", *CameraId);
		return true;
	}
	catch (...)
	{
		MIXCAST_LOG("Failed to take snapshot for camera: ", *CameraId);
		return false;
	}
}

void FServiceCommunicator::NotifySdkStarted()
{
	PopulateExpData();

	//MIXCAST_LOG("Tracking Type is %d", expData.trackingType);
	try
	{
		SDKServiceClient.Run<void>([this](SDK_ServiceClient& Client)
		{
			Client.NotifySdkStarted(expData);
		});

		//MIXCAST_LOG("Successfully notified Service that SDK has started");
	}
	catch (...)
	{
		//MIXCAST_LOG("Failed to notify Service that SDK has started");
	}
}

void FServiceCommunicator::NotifySdkStopped() const
{
	try
	{
		SDKServiceClient.Run<void>([this](SDK_ServiceClient& Client)
		{
			Client.NotifySdkStopped();
		});

		//MIXCAST_LOG("Successfully notified Service that SDK has stopped");
	}
	catch (...)
	{
		//MIXCAST_LOG("Failed to notify Service that SDK has stopped");
	}
}

void FServiceCommunicator::NotifySdkChanged() const
{
	//MIXCAST_LOG("Tracking Type is %d", expData.trackingType);
	try
	{
		SDKServiceClient.Run<void>([this](SDK_ServiceClient& Client)
			{
				Client.NotifySdkChanged(expData);
			});

		//MIXCAST_LOG("Successfully notified Service that SDK has been updated");
	}
	catch (...)
	{
		//MIXCAST_LOG("Failed to notify Service that SDK has updated");
	}
}

SharedTex FServiceCommunicator::GetSharedTextureInfo(const FString& TextureId)
{
	try {
		return SharedTextureClient.Run<SharedTex>([&TextureId](SharedTextureCommunicationClient& Client)
			{
				SharedTex result;
				Client.SharedTextureRequest(result, TCHAR_TO_UTF8(*TextureId));
				return result;
			});
	}
	catch (...)
	{
		MIXCAST_ERROR("Failed to get shared texture with ID %s", *TextureId);
		return SharedTex();
	}
}

void FServiceCommunicator::UpdateExperienceTrackedObjects(const std::vector<TrackedObject>& objs)
{
	try
	{
		SDKServiceClient.Run<void>([&](SDK_ServiceClient& Client)
			{
				Client.UpdateExperienceTrackedObjectMetadata(objs);
			});

		//MIXCAST_LOG("Successfully notified Service of custom objects");
	}
	catch (...)
	{
		//MIXCAST_LOG("Failed to notify Service that SDK of custom objects");
	}
}

void FServiceCommunicator::HandleActiveChanged()
{
	if (IMixCast::IsAvailable() && IMixCast::Get().IsActive())
	{
		if (ServiceMutexCheck == nullptr)
			ServiceMutexCheck = new FServiceMutexThread();
	}
	else
	{
		if (ServiceMutexCheck) {
			delete ServiceMutexCheck;
			ServiceMutexCheck = nullptr;
		}

		SharedTextureQueue.Empty();
		SharedTextureQueueSize = 0;

		SDKServiceClient.EnsureTransportIsClosed();
		SharedTextureClient.EnsureTransportIsClosed();
	}
}

void FServiceCommunicator::PopulateExpData()
{
	expData.mixcastVersion = StringUtils::ToStdString(FMixCast::GetPluginVersion());

	expData.experienceExePath = StringUtils::ToStdString(ProjectUtils::GetExecutablePath());
	expData.mainWindowHandle = ProjectUtils::GetMainWindowHandle();
	expData.mainProcessId = FPlatformProcess::GetCurrentProcessId();

	expData.projectId = StringUtils::ToStdString(ProjectUtils::GetProjectID());
	expData.experienceTitle = StringUtils::ToStdString(ProjectUtils::GetProjectName());
	expData.organizationName = StringUtils::ToStdString(ProjectUtils::GetCompanyName());

	expData.engineVersion = StringUtils::ToStdString(ProjectUtils::GetEngineVersion());
	expData.graphicsType = ProjectUtils::GetGraphicsType();
	expData.trackingType = TrackingUtil::GetCurrentTrackingType(false);

	expData.alphaIsPremultiplied = true;
	expData.colorSpaceIsLinear = true;

	UPluginProjectSettings* projectSettings = SdkData::GetProjectSettings();
	expData.canRenderOpaqueBg = projectSettings->CanRenderOpaqueBg;
	expData.canRenderTransparentBg = projectSettings->CanRenderTransparentBg;
#if MIXCAST_SHADER_PASSES
	expData.canRenderSeparateOpaque = !projectSettings->DoesntProduceGoodDepth;
#else
	expData.canRenderSeparateOpaque = false;
#endif
}