/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "ReadOpaque.h"
#if MIXCAST_SHADER_PASSES
#include "CommonRenderResources.h"
#include "Containers/DynamicRHIResourceArray.h"
#include "Runtime/Renderer/Private/PostProcess/SceneRenderTargets.h"

class FReadOpaqueVertexBuffer : public FVertexBuffer
{
public:
	/** Initialize the RHI for this rendering resource */
	void InitRHI()
	{
		TResourceArray<FFilterVertex, VERTEXBUFFER_ALIGNMENT> Vertices;
		Vertices.SetNumUninitialized(6);

		Vertices[0].Position = FVector4(-1, 1, 0, 1);
		Vertices[0].UV = FVector2D(0, 0);

		Vertices[1].Position = FVector4(1, 1, 0, 1);
		Vertices[1].UV = FVector2D(1, 0);

		Vertices[2].Position = FVector4(-1, -1, 0, 1);
		Vertices[2].UV = FVector2D(0, 1);

		Vertices[3].Position = FVector4(1, -1, 0, 1);
		Vertices[3].UV = FVector2D(1, 1);

		// Create vertex buffer. Fill buffer with initial data upon creation
		FRHIResourceCreateInfo CreateInfo(&Vertices);
		VertexBufferRHI = RHICreateVertexBuffer(Vertices.GetResourceDataSize(), BUF_Static, CreateInfo);
	}
};
TGlobalResource<FReadOpaqueVertexBuffer> GReadOpaqueVertexBuffer;

class FReadOpaqueVS : public FGlobalShader
{
	DECLARE_GLOBAL_SHADER(FReadOpaqueVS);

public:
	FReadOpaqueVS() { }
	FReadOpaqueVS(const ShaderMetaType::CompiledShaderInitializerType& Initializer)
		: FGlobalShader(Initializer)
	{
		BufferRatioParam.Bind(Initializer.ParameterMap, TEXT("BufferRatio"));
	}

#if !UE_4_25_OR_LATER
	virtual bool Serialize(FArchive& Ar) override
	{
		bool bShaderHasOutdatedParameters = FGlobalShader::Serialize(Ar);
		Ar << BufferRatioParam;
		return bShaderHasOutdatedParameters;
	}
#endif

	static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
	{
		return true;
	}

protected:
#if UE_4_25_OR_LATER
	LAYOUT_FIELD(FShaderParameter, BufferRatioParam);
#else
	FShaderParameter BufferRatioParam;
#endif

public:

	void SetParameters(FRHICommandList& RHICmdList, TShaderMapRef<FReadOpaqueVS> ShaderRef, FSceneView& InView, FVector4 BufferRatio)
	{
#if !UE_4_25_OR_LATER
		FRHIVertexShader* ShaderRHI = GetVertexShader();
#else
		FRHIVertexShader* ShaderRHI = ShaderRef.GetVertexShader();
#endif

		FGlobalShader::SetParameters<FViewUniformShaderParameters>(RHICmdList, ShaderRHI, InView.ViewUniformBuffer);
		SetShaderValue(RHICmdList, ShaderRHI, BufferRatioParam, BufferRatio);
	}
};

template<bool ReadDepth>
class FReadOpaquePS : public FGlobalShader
{
public:
	DECLARE_GLOBAL_SHADER(FReadOpaquePS);

	static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
	{
		return IsFeatureLevelSupported(Parameters.Platform, ERHIFeatureLevel::SM5);
	}

	static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
	{
		FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
		OutEnvironment.SetDefine(TEXT("READ_DEPTH"), ReadDepth ? 1 : 0);
	}

protected:
#if UE_4_25_OR_LATER
	LAYOUT_FIELD(FShaderResourceParameter, DepthTexture);
	LAYOUT_FIELD(FShaderResourceParameter, DepthTextureSampler);

	LAYOUT_FIELD(FShaderResourceParameter, ColorTexture);
	LAYOUT_FIELD(FShaderResourceParameter, ColorTextureSampler);
#else
	FShaderResourceParameter DepthTexture;
	FShaderResourceParameter DepthTextureSampler;

	FShaderResourceParameter ColorTexture;
	FShaderResourceParameter ColorTextureSampler;
#endif

public:
	FReadOpaquePS() { }
	FReadOpaquePS(const ShaderMetaType::CompiledShaderInitializerType& Initializer)
		: FGlobalShader(Initializer)
	{
		DepthTexture.Bind(Initializer.ParameterMap, TEXT("DepthTexture"));
		DepthTextureSampler.Bind(Initializer.ParameterMap, TEXT("DepthTextureSampler"));

		ColorTexture.Bind(Initializer.ParameterMap, TEXT("ColorTexture"));
		ColorTextureSampler.Bind(Initializer.ParameterMap, TEXT("ColorTextureSampler"));
	}

#if !UE_4_25_OR_LATER
	virtual bool Serialize(FArchive& Ar) override
	{
		bool bShaderHasOutdatedParameters = FGlobalShader::Serialize(Ar);
		Ar << DepthTexture << DepthTextureSampler << ColorTexture << ColorTextureSampler;
		return bShaderHasOutdatedParameters;
	}
#endif

	 void SetParameters(FRHICommandList& RHICmdList, TShaderMapRef<FReadOpaquePS> ShaderRef, FSceneView& InView, FRHITexture* ColorTextureRHI, FRHITexture* DepthTextureRHI)
	 {
		 FRHISamplerState* SamplerStateRHI = TStaticSamplerState<SF_Bilinear>::GetRHI();
 #if !UE_4_25_OR_LATER
		 FRHIPixelShader* ShaderRHI = GetPixelShader();
 #else
		 FRHIPixelShader* ShaderRHI = ShaderRef.GetPixelShader();
 #endif

		 FGlobalShader::SetParameters<FViewUniformShaderParameters>(RHICmdList, ShaderRHI, InView.ViewUniformBuffer); 
		 SetTextureParameter(RHICmdList, ShaderRHI, ColorTexture, ColorTextureSampler, SamplerStateRHI, ColorTextureRHI);
		 SetTextureParameter(RHICmdList, ShaderRHI, DepthTexture, DepthTextureSampler, SamplerStateRHI, DepthTextureRHI);
	 }
};

IMPLEMENT_GLOBAL_SHADER(FReadOpaqueVS, "/Plugin/MixCast/ReadOpaque.usf", "MainVS", SF_Vertex);

IMPLEMENT_GLOBAL_SHADER(FReadOpaquePS<false>,	"/Plugin/MixCast/ReadOpaque.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FReadOpaquePS<true>,	"/Plugin/MixCast/ReadOpaque.usf", "MainPS", SF_Pixel);


void FMixCastShader_ReadOpaque::Execute(IRendererModule& RendererModule, FPostOpaqueRenderParameters& Parameters, UTextureRenderTarget2D& OpaqueLayerTarget, bool ReadDepth)
{
	FSceneRenderTargets& SceneContext = FSceneRenderTargets::Get(*Parameters.RHICmdList);
	FSceneView* InView = (FSceneView*)Parameters.Uid;


	if (Parameters.RHICmdList->IsInsideRenderPass())
		Parameters.RHICmdList->EndRenderPass();

	FRHITexture* targets[] = { OpaqueLayerTarget.Resource->TextureRHI };
	FRHIRenderPassInfo OpaqueGrabPassInfo(1, targets, ERenderTargetActions::DontLoad_Store);
	Parameters.RHICmdList->BeginRenderPass(OpaqueGrabPassInfo, TEXT("MixCast-ReadOpaque"));

	auto ShaderMap = GetGlobalShaderMap(InView->GetFeatureLevel());

	TShaderMapRef<FReadOpaqueVS> VertexShader(ShaderMap);

	//prepare state
	FGraphicsPipelineStateInitializer GraphicsPSOInit;
	Parameters.RHICmdList->ApplyCachedRenderTargets(GraphicsPSOInit);

	GraphicsPSOInit.BlendState = TStaticBlendState<>::GetRHI();
	GraphicsPSOInit.RasterizerState = TStaticRasterizerState<>::GetRHI();
	GraphicsPSOInit.DepthStencilState = TStaticDepthStencilState<false, CF_Always>::GetRHI();

	GraphicsPSOInit.BoundShaderState.VertexDeclarationRHI = GFilterVertexDeclaration.VertexDeclarationRHI;
	GraphicsPSOInit.PrimitiveType = PT_TriangleStrip;
#if UE_4_25_OR_LATER
	GraphicsPSOInit.BoundShaderState.VertexShaderRHI = VertexShader.GetVertexShader();
#else
	GraphicsPSOInit.BoundShaderState.VertexShaderRHI = GETSAFERHISHADER_VERTEX(*VertexShader);
#endif

	if (ReadDepth) 
	{
		TShaderMapRef<FReadOpaquePS<true>> MRTPixelShader(ShaderMap);
#if UE_4_25_OR_LATER
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = MRTPixelShader.GetPixelShader();
#else
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = GETSAFERHISHADER_PIXEL(*MRTPixelShader);
#endif
		SetGraphicsPipelineState(*Parameters.RHICmdList, GraphicsPSOInit);

		MRTPixelShader->SetParameters(*Parameters.RHICmdList, MRTPixelShader, *InView,
			SceneContext.GetSceneColor()->GetRenderTargetItem().ShaderResourceTexture, Parameters.DepthTexture);
	}
	else
	{
		TShaderMapRef<FReadOpaquePS<false>> MRTPixelShader(ShaderMap);
#if UE_4_25_OR_LATER
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = MRTPixelShader.GetPixelShader();
#else
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = GETSAFERHISHADER_PIXEL(*MRTPixelShader);
#endif
		SetGraphicsPipelineState(*Parameters.RHICmdList, GraphicsPSOInit);

		MRTPixelShader->SetParameters(*Parameters.RHICmdList, MRTPixelShader, *InView,
			SceneContext.GetSceneColor()->GetRenderTargetItem().ShaderResourceTexture, NULL);
	}

	FPooledRenderTargetDesc Desc = SceneContext.GetSceneColor()->GetDesc();
	FVector4 bufferRatio(OpaqueLayerTarget.GetSurfaceWidth() / Desc.Extent.X,
		OpaqueLayerTarget.GetSurfaceHeight() / Desc.Extent.Y);

	VertexShader->SetParameters(*Parameters.RHICmdList, VertexShader, *InView,
		bufferRatio);


	Parameters.RHICmdList->SetViewport(0, 0, 0, OpaqueLayerTarget.GetSurfaceWidth(), OpaqueLayerTarget.GetSurfaceHeight(), 1);
	Parameters.RHICmdList->SetStreamSource(0, GReadOpaqueVertexBuffer.VertexBufferRHI, 0);
	Parameters.RHICmdList->DrawPrimitive(0, 2, 1);

	Parameters.RHICmdList->EndRenderPass();

#if !UE_4_26_OR_LATER
#if UE_4_25_OR_LATER
	FRHIRenderPassInfo PostOpaqueGrabInfo;
	int32 velIndex = -1;
	int32 tangentIndex = -1;
	SceneContext.FillGBufferRenderPassInfo(ERenderTargetLoadAction::ELoad, PostOpaqueGrabInfo, velIndex, tangentIndex);

	PostOpaqueGrabInfo.DepthStencilRenderTarget.DepthStencilTarget = SceneContext.GetSceneDepthSurface();
	PostOpaqueGrabInfo.DepthStencilRenderTarget.ExclusiveDepthStencil = FExclusiveDepthStencil::DepthWrite_StencilNop;
	PostOpaqueGrabInfo.DepthStencilRenderTarget.Action = MakeDepthStencilTargetActions(ERenderTargetActions::Load_Store, ERenderTargetActions::DontLoad_DontStore);
	Parameters.RHICmdList->BeginRenderPass(PostOpaqueGrabInfo, TEXT("Post-MixCast-ReadOpaqueColor"));
#else
	SceneContext.BeginRenderingGBuffer(*Parameters.RHICmdList, ERenderTargetLoadAction::ELoad, ERenderTargetLoadAction::ELoad, FExclusiveDepthStencil::DepthNop_StencilNop, false);
#endif
#endif
}
#endif